@@ -49,6 +49,41 @@ declaration: Function - root
       )"},
       {R"cpp(
 namespace root {
+struct S { static const int x = 0; ~S(); };
+int y = S::x + root::S().x;
+}
+      )cpp",
+       R"(
+declaration: Namespace - root
+  declaration: CXXRecord - S
+    declaration: Var - x
+      type: Qualified - const
+        type: Builtin - int
+      expression: IntegerLiteral - 0
+    declaration: CXXDestructor
+      type: Record - S
+      type: FunctionProto
+        type: Builtin - void
+    declaration: CXXConstructor
+    declaration: CXXConstructor
+  declaration: Var - y
+    type: Builtin - int
+    expression: ExprWithCleanups
+      expression: BinaryOperator - +
+        expression: ImplicitCast - LValueToRValue
+          expression: DeclRef - x
+            specifier: TypeSpec
+              type: Record - S
+        expression: ImplicitCast - LValueToRValue
+          expression: Member - x
+            expression: CXXBindTemporary
+              expression: CXXTemporaryObject - S
+                type: Elaborated
+                  specifier: Namespace - root::
+                  type: Record - S
+      )"},
+      {R"cpp(
+namespace root {
 struct S { static const int x = 0; };
 int y = S::x + root::S().x;
 }
@@ -66,19 +101,17 @@ declaration: Namespace - root
     declaration: CXXDestructor
   declaration: Var - y
     type: Builtin - int
-    expression: ExprWithCleanups
-      expression: BinaryOperator - +
-        expression: ImplicitCast - LValueToRValue
-          expression: DeclRef - x
-            specifier: TypeSpec
+    expression: BinaryOperator - +
+      expression: ImplicitCast - LValueToRValue
+        expression: DeclRef - x
+          specifier: TypeSpec
+            type: Record - S
+      expression: ImplicitCast - LValueToRValue
+        expression: Member - x
+          expression: CXXTemporaryObject - S
+            type: Elaborated
+              specifier: Namespace - root::
               type: Record - S
-        expression: ImplicitCast - LValueToRValue
-          expression: Member - x
-            expression: MaterializeTemporary - rvalue
-              expression: CXXTemporaryObject - S
-                type: Elaborated
-                  specifier: Namespace - root::
-                  type: Record - S
       )"},
       {R"cpp(
 namespace root {
